"use strict";
/**
 * @license
 * Copyright 2018 Google Inc. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */
Object.defineProperty(exports, "__esModule", { value: true });
var tf = require("../index");
var jasmine_util_1 = require("../jasmine_util");
var test_util_1 = require("../test_util");
var optimizer_1 = require("./optimizer");
var sgd_optimizer_1 = require("./sgd_optimizer");
jasmine_util_1.describeWithFlags('optimizer', test_util_1.ALL_ENVS, function () {
    it('basic', function () {
        var learningRate = .1;
        var optimizer = tf.train.sgd(learningRate);
        var x = tf.scalar(4).variable();
        var bias = tf.scalar(1).variable();
        var strayVariable = tf.scalar(-1).variable();
        var numTensors = tf.memory().numTensors;
        var f = function () { return x.square().addStrict(bias); };
        var cost = optimizer.minimize(f, /* returnCost */ true);
        // Cost should be the only additional array.
        expect(tf.memory().numTensors).toBe(numTensors + 1);
        // de/dx = 2x
        var expectedX1 = -2 * 4 * learningRate + 4;
        // de/db = 1
        var expectedBias1 = -1 * learningRate + 1;
        test_util_1.expectArraysClose(x, [expectedX1]);
        test_util_1.expectArraysClose(bias, [expectedBias1]);
        test_util_1.expectArraysClose(cost, [Math.pow(4, 2) + 1]);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
        cost.dispose();
        numTensors = tf.memory().numTensors;
        cost = optimizer.minimize(f, /* returnCost */ false);
        // There should be no new additional Tensors.
        expect(tf.memory().numTensors).toBe(numTensors);
        var expectedX2 = -2 * expectedX1 * learningRate + expectedX1;
        var expectedBias2 = -learningRate + expectedBias1;
        test_util_1.expectArraysClose(x, [expectedX2]);
        test_util_1.expectArraysClose(bias, [expectedBias2]);
        expect(cost).toBe(null);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
        optimizer.dispose();
        x.dispose();
        bias.dispose();
        strayVariable.dispose();
        // The only tensors remaining are the arguments to variable().
        expect(tf.memory().numTensors).toBe(3);
    });
    it('varList array of all variables', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        var x = tf.scalar(4).variable();
        var bias = tf.scalar(1).variable();
        var strayVariable = tf.scalar(-1).variable();
        var varList = [x, bias];
        var f = function () { return x.square().addStrict(bias); };
        var cost = optimizer.minimize(f, /* returnCost */ true, varList);
        // de/dx = 2x
        var expectedX1 = -2 * 4 * learningRate + 4;
        // de/db = 1
        var expectedBias1 = -1 * learningRate + 1;
        test_util_1.expectArraysClose(x, [expectedX1]);
        test_util_1.expectArraysClose(bias, [expectedBias1]);
        test_util_1.expectArraysClose(cost, [Math.pow(4, 2) + 1]);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
        cost = optimizer.minimize(f, /* returnCost */ false, varList);
        var expectedX2 = -2 * expectedX1 * learningRate + expectedX1;
        var expectedBias2 = -learningRate + expectedBias1;
        test_util_1.expectArraysClose(x, [expectedX2]);
        test_util_1.expectArraysClose(bias, [expectedBias2]);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
        expect(cost).toBe(null);
    });
    it('varList empty array of variables throws error', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        var x = tf.scalar(4).variable();
        var bias = tf.scalar(1).variable();
        // Stray variable.
        tf.scalar(-1).variable();
        var varList = [];
        var f = function () { return x.square().addStrict(bias); };
        expect(function () { return optimizer.minimize(f, /* returnCost */ true, varList); })
            .toThrowError();
    });
    it('varList subset of variables update', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        var x = tf.scalar(4).variable();
        var bias = tf.scalar(1).variable();
        var strayVariable = tf.scalar(-1).variable();
        var varList = [x];
        var f = function () { return x.square().addStrict(bias); };
        var cost = optimizer.minimize(f, /* returnCost */ true, varList);
        // de/dx = 2x
        var expectedValue1 = -2 * 4 * learningRate + 4;
        test_util_1.expectArraysClose(x, [expectedValue1]);
        // bias should remain unchanged.
        test_util_1.expectArraysClose(bias, [1]);
        test_util_1.expectArraysClose(cost, [Math.pow(4, 2) + 1]);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
        cost = optimizer.minimize(f, /* returnCost */ false, varList);
        var expectedValue2 = -2 * expectedValue1 * learningRate + expectedValue1;
        test_util_1.expectArraysClose(x, [expectedValue2]);
        // Bias still should remain unchanged.
        test_util_1.expectArraysClose(bias, [1]);
        expect(cost).toBe(null);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
    });
    it('only bias trainable', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        var trainable = false;
        var x = tf.scalar(4).variable(trainable);
        var bias = tf.scalar(1).variable();
        var strayVariable = tf.scalar(-1).variable();
        var f = function () { return x.square().addStrict(bias); };
        var cost = optimizer.minimize(f, /* returnCost */ true);
        // x should not have been updated.
        test_util_1.expectArraysClose(x, [4]);
        // de/db = 1
        var expectedBias1 = -1 * learningRate + 1;
        test_util_1.expectArraysClose(bias, [expectedBias1]);
        test_util_1.expectArraysClose(cost, [Math.pow(4, 2) + 1]);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
        cost = optimizer.minimize(f, /* returnCost */ false);
        // x should not have been updated.
        test_util_1.expectArraysClose(x, [4]);
        var expectedBias2 = -learningRate + expectedBias1;
        test_util_1.expectArraysClose(bias, [expectedBias2]);
        expect(cost).toBe(null);
        // The stray variable should remain unchanged.
        test_util_1.expectArraysClose(strayVariable, [-1]);
    });
    it('only bias trainable, only x in varList throws error', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        var trainable = false;
        var x = tf.scalar(4).variable(trainable);
        var bias = tf.scalar(1).variable();
        // stray variable.
        tf.scalar(-1).variable();
        var varList = [x];
        var f = function () { return x.square().addStrict(bias); };
        expect(function () { return optimizer.minimize(f, /* returnCost */ true, varList); })
            .toThrowError();
    });
    it('instanceof Optimizer', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        expect(optimizer instanceof optimizer_1.Optimizer).toBe(true);
    });
    it('throws error when f returns a non-scalar', function () {
        var learningRate = .1;
        var optimizer = new sgd_optimizer_1.SGDOptimizer(learningRate);
        var x = tf.tensor1d([1, 2]).variable();
        var f = function () { return x.square(); };
        // tslint:disable-next-line:no-any
        expect(function () { return optimizer.minimize(f); }).toThrowError();
    });
});
//# sourceMappingURL=optimizer_test.js.map